# -*- coding: utf-8 -*-
import numpy as np
import onnxruntime
import cv2
import onnx
from onnx import numpy_helper

from face2face.core.compatibility import face_align


class INSwapper:
    def __init__(self, model_file=None, providers=None):
        self.model_file = model_file
        model = onnx.load(self.model_file)
        graph = model.graph
        self.emap = numpy_helper.to_array(graph.initializer[-1])
        self.input_mean = 0.0
        self.input_std = 255.0
        self.session = onnxruntime.InferenceSession(self.model_file, providers=providers)
        inputs = self.session.get_inputs()
        self.input_names = []
        for inp in inputs:
            self.input_names.append(inp.name)
        outputs = self.session.get_outputs()
        output_names = []
        for out in outputs:
            output_names.append(out.name)
        self.output_names = output_names
        assert len(self.output_names)==1
        output_shape = outputs[0].shape
        input_cfg = inputs[0]
        input_shape = input_cfg.shape
        self.input_shape = input_shape
        print('inswapper-shape:', self.input_shape)
        self.input_size = tuple(input_shape[2:4][::-1])

    def forward(self, img, latent):
        img = (img - self.input_mean) / self.input_std
        pred = self.session.run(self.output_names, {self.input_names[0]: img, self.input_names[1]: latent})[0]
        return pred

    def get(self, img, target_face, source_face, paste_back=True):
        aimg, M = face_align.norm_crop2(img, target_face.kps, self.input_size[0])
        blob = cv2.dnn.blobFromImage(aimg, 1.0 / self.input_std, self.input_size,
                                      (self.input_mean, self.input_mean, self.input_mean), swapRB=True)
        latent = source_face.normed_embedding.reshape((1,-1))
        latent = np.dot(latent, self.emap)
        latent /= np.linalg.norm(latent)
        pred = self.session.run(self.output_names, {self.input_names[0]: blob, self.input_names[1]: latent})[0]
        #print(latent.shape, latent.dtype, pred.shape)
        img_fake = pred.transpose((0,2,3,1))[0]
        bgr_fake = np.clip(255 * img_fake, 0, 255).astype(np.uint8)[:,:,::-1]
        if not paste_back:
            return bgr_fake, M
        else:
            target_img = img
            fake_diff = bgr_fake.astype(np.float32) - aimg.astype(np.float32)
            fake_diff = np.abs(fake_diff).mean(axis=2)
            fake_diff[:2,:] = 0
            fake_diff[-2:,:] = 0
            fake_diff[:,:2] = 0
            fake_diff[:,-2:] = 0
            IM = cv2.invertAffineTransform(M)
            img_white = np.full((aimg.shape[0],aimg.shape[1]), 255, dtype=np.float32)
            bgr_fake = cv2.warpAffine(bgr_fake, IM, (target_img.shape[1], target_img.shape[0]), borderValue=0.0)
            img_white = cv2.warpAffine(img_white, IM, (target_img.shape[1], target_img.shape[0]), borderValue=0.0)
            fake_diff = cv2.warpAffine(fake_diff, IM, (target_img.shape[1], target_img.shape[0]), borderValue=0.0)
            img_white[img_white>20] = 255
            fthresh = 10
            fake_diff[fake_diff<fthresh] = 0
            fake_diff[fake_diff>=fthresh] = 255
            img_mask = img_white
            mask_h_inds, mask_w_inds = np.where(img_mask==255)
            mask_h = np.max(mask_h_inds) - np.min(mask_h_inds)
            mask_w = np.max(mask_w_inds) - np.min(mask_w_inds)
            mask_size = int(np.sqrt(mask_h*mask_w))
            k = max(mask_size//10, 10)
            #k = max(mask_size//20, 6)
            #k = 6
            kernel = np.ones((k,k),np.uint8)
            img_mask = cv2.erode(img_mask,kernel,iterations = 1)
            kernel = np.ones((2,2),np.uint8)
            fake_diff = cv2.dilate(fake_diff,kernel,iterations = 1)
            k = max(mask_size//20, 5)
            #k = 3
            #k = 3
            kernel_size = (k, k)
            blur_size = tuple(2*i+1 for i in kernel_size)
            img_mask = cv2.GaussianBlur(img_mask, blur_size, 0)
            k = 5
            kernel_size = (k, k)
            blur_size = tuple(2*i+1 for i in kernel_size)
            fake_diff = cv2.GaussianBlur(fake_diff, blur_size, 0)
            img_mask /= 255
            fake_diff /= 255
            #img_mask = fake_diff
            img_mask = np.reshape(img_mask, [img_mask.shape[0],img_mask.shape[1],1])
            fake_merged = img_mask * bgr_fake + (1-img_mask) * target_img.astype(np.float32)
            fake_merged = fake_merged.astype(np.uint8)
            return fake_merged





# import cv2
# import numpy

# from facefusion.globals import conditional_thread_semaphore
# from facefusion.vision_frame import VisionFrame
# from facefusion.mask import Mask


# def create_occlusion_mask(crop_vision_frame: VisionFrame) -> Mask:
#     face_occluder = get_face_occluder()
#     prepare_vision_frame = cv2.resize(crop_vision_frame, face_occluder.get_inputs()[0].shape[1:3][::-1])
#     prepare_vision_frame = numpy.expand_dims(prepare_vision_frame, axis=0).astype(numpy.float32) / 255
#     prepare_vision_frame = prepare_vision_frame.transpose(0, 1, 2, 3)
#     with conditional_thread_semaphore(facefusion.globals.execution_providers):
#         occlusion_mask: Mask = face_occluder.run(None,
#                                                  {
#                                                      face_occluder.get_inputs()[0].name: prepare_vision_frame
#                                                  })[0][0]
#     occlusion_mask = occlusion_mask.transpose(0, 1, 2).clip(0, 1).astype(numpy.float32)
#     occlusion_mask = cv2.resize(occlusion_mask, crop_vision_frame.shape[:2][::-1])
#     occlusion_mask = (cv2.GaussianBlur(occlusion_mask.clip(0, 1), (0, 0), 5).clip(0.5, 1) - 0.5) * 2
#     return occlusion_mask



# avoid circular dependency but provide type hints
from __future__ import annotations
from typing import TYPE_CHECKING, Union, List, Tuple, Dict

from face2face.core.compatibility.Face import Face
from media_toolkit import ImageFile, MediaFile, MediaList

if TYPE_CHECKING:
    from face2face.core.face2face import Face2Face

# regular imports
import glob
import os

import numpy as np


from face2face.core.modules.storage.f2f_loader import load_reference_face_from_file
from face2face.core.modules.storage.file_writable_face import FileWriteableFace
from face2face.settings import EMBEDDINGS_DIR
from face2face.core.modules.utils.utils import encode_path_safe
from face2face.core.modules.utils.utils import load_image


class _FaceEmbedding:
    def load_face(self: Face2Face, face_name: str) -> Face:
        """
        Load a reference face embedding from a file.
        :param face_name: the name of the reference face embedding
        :return: the embedding of the reference face(s)
        """
        # check if is already in ram. If yes return that one
        embedding = self._face_embeddings.get(face_name, None)
        if embedding is not None:
            return embedding

        # load from file
        file = os.path.join(self._face_embedding_folder, f"{face_name}.npy")
        face = load_reference_face_from_file(file)

        # add to memory dict
        self._face_embeddings[face_name] = face
        return face

    def load_faces(
        self, 
        face_names: Union[str, List[str], List[Face], MediaFile, MediaList, None] = None
    ) -> Dict[str, Face]:
        """
        Load face embeddings from various sources.
        
        Args:
            face_names: The source of face embeddings to load. Can be:
                - None: Loads all stored face embeddings from the _face_embeddings folder
                - str: Loads a single face embedding by name
                - List[str]: Loads multiple face embeddings by their names
                - List[Face]: Returns the faces as {index: face}
                - MediaFile: Loads a face from the file content (e.g., from API response)
                - MediaList: Loads faces from each file in the list, supporting both MediaFile and string paths
        
        Returns:
            Dict[str, Face]: Dictionary mapping face names to their embeddings.
                For MediaFile inputs, uses the file name as the key.
                For MediaList inputs, uses file names or string paths as keys.
                For List[Face] inputs, uses numeric indices as keys.
        
        Raises:
            ValueError: If a face file is not found or cannot be loaded
        """
        if face_names is None:
            return self.load_all_faces()
        elif isinstance(face_names, str):
            return {face_names: self.load_face(face_names)}
        elif isinstance(face_names, MediaFile):
            # Load face from MediaFile content
            face = load_reference_face_from_file(face_names.to_bytes_io())
            return {face_names.file_name: face}
        elif isinstance(face_names, MediaList):
            # Load faces from each file in MediaList
            ret = {}
            for i, file in enumerate(face_names):
                if isinstance(file, MediaFile):
                    face = load_reference_face_from_file(file.to_bytes_io())
                    ret[file.file_name] = face
                elif isinstance(file, str):
                    ret[file] = self.load_face(file)
            return ret

        # convert whatever list to dict
        ret = {}
        for i, face in enumerate(face_names):
            if isinstance(face, Face):
                ret[i] = face
            elif isinstance(face, str):
                ret[face] = self.load_face(face)

        return ret

    def load_all_faces(self: Face2Face) -> Dict[str, Face]:
        """
        Load all face embeddings from the _face_embeddings folder.
        
        Returns:
            Dict[str, Face]: Dictionary mapping face names to their embeddings
        """
        for face_name in glob.glob(self._face_embedding_folder + "/*.npy"):
            self.load_face(face_name)
        return self._face_embeddings

    def add_face(
        self: Face2Face,
        face_name: Union[str, List[str]],
        image: Union[np.array, str, ImageFile],
        save: bool = False
    ) -> Union[Tuple[str, FileWriteableFace], Dict[str, Face]]:
        """
        Add one or multiple reference face(s) to the face swapper. This face(s) can be used for swapping in other images.

        Args:
            face_name: The name(s) for the reference face(s).
                - If a single string, creates one face embedding
                - If a list of strings, creates embeddings for each face from left to right in the image
            image: The image from which to extract the face(s). Can be:
                - numpy array: Direct image data
                - str: Path to image file
                - ImageFile: MediaFile containing the image
            save: If True, saves the face embeddings to disk in the _face_embeddings folder.
                If False, only stores them in memory.

        Returns:
            Union[Tuple[str, FileWriteableFace], Dict[str, Face]]:
                - For single face: Tuple of (face_name, FileWriteableFace)
                - For multiple faces: Dict mapping face names to their embeddings

        Raises:
            ValueError: If no face name is provided or no faces are detected in the image
        """
        if isinstance(face_name, list) and len(face_name) == 0 or face_name is None:
            raise ValueError("Please provide at least one face name.")

        try:
            image = load_image(image)
            detected_faces = self.detect_faces(image)
        except Exception as e:
            raise ValueError(f"Could not load image or face detection failed. Error: {e}")

        # Deal with errors of too much or too few faces in the reference image
        if not detected_faces:
            raise ValueError(f"No faces detected in the provided image for {face_name}.")

        if not isinstance(face_name, list):
            face_name = [face_name]

        if len(face_name) > len(detected_faces):
            print(f"Not enough faces in the image for all provided face names. "
                  f"Only {len(detected_faces)} faces found. Skipping the rest")
            face_name = face_name[:len(detected_faces)]

        face = None
        for i, name in enumerate(face_name):
            name = encode_path_safe(name)
            face = detected_faces[i]
            # Store the detected faces in memory
            self._face_embeddings[name] = face

            # store the detected faces on disc
            face = FileWriteableFace(face)
            # Save face to virtual file
            if save:
                os.makedirs(EMBEDDINGS_DIR, exist_ok=True)
                filename = os.path.join(EMBEDDINGS_DIR, f"{name}.npy")
                if os.path.isfile(filename):
                    print(f"Reference face {name} already exists. Overwriting.")
                face.to_file(filename)
        if len(face_name) > 1:
            # Return a dict with the face names and embeddings
            return {
                name: self._face_embeddings[name]
                for name in face_name
            }
        else:
            return face_name[0], face



from face2face.settings import MODELS_DIR


SWAPPER_MODELS = {
    "inswapper_128": {
        'url': "https://socaityfiles.blob.core.windows.net/model-zoo/face2face/insightface/inswapper_128.onnx",
        'path': f'{MODELS_DIR}/insightface/inswapper_128.onnx',
    }
}

INSIGHT_FACE_MODELS = {
    "buffalo_l": {
        'url': "https://socaityfiles.blob.core.windows.net/model-zoo/face2face/insightface/buffalo_l.zip",
        # https://github.com/deepinsight/insightface/releases/download/v0.7/buffalo_l.zip
        'path': f'{MODELS_DIR}/insightface/buffalo_l.zip',  # gets unpacked and downloaded to buffalo_l/*
    }
}

FACE_ENHANCER_MODELS = {
    'gfpgan_1.4': {
            'url': 'https://socaityfiles.blob.core.windows.net/model-zoo/face2face/face_enhancer/gfpgan_1.4.onnx',
            'path': f'{MODELS_DIR}/face_enhancer/gfpgan_1.4.onnx',
            'template': 'ffhq_512',
            'size': (512, 512)
    },
    'gpen_bfr_256':{
            'url': 'https://socaityfiles.blob.core.windows.net/model-zoo/face2face/face_enhancer/gpen_bfr_256.onnx',
            'path': f'{MODELS_DIR}/face_enhancer/gpen_bfr_256.onnx',
            'template': 'arcface_128_v2',
            'size': (256, 256)
    },
    'gpen_bfr_512': {
            'url': 'https://socaityfiles.blob.core.windows.net/model-zoo/face2face/face_enhancer/gpen_bfr_512.onnx',
            'path': f'{MODELS_DIR}/face_enhancer/gpen_bfr_512.onnx',
            'template': 'ffhq_512',
            'size': (512, 512)
    },
    'gpen_bfr_1024': {
            'url': 'https://socaityfiles.blob.core.windows.net/model-zoo/face2face/face_enhancer/gpen_bfr_1024.onnx',
            'path': f'{MODELS_DIR}/face_enhancer/gpen_bfr_1024.onnx',
            'template': 'ffhq_512',
            'size': (1024, 1024)
    },
    'gpen_bfr_2048': {
            'url': 'https://socaityfiles.blob.core.windows.net/model-zoo/face2face/face_enhancer/gpen_bfr_2048.onnx',
            'path': f'{MODELS_DIR}/face_enhancer/gpen_bfr_2048.onnx',
            'template': 'ffhq_512',
            'size': (2048, 2048)
    }
}

OCCLUSION_MODELS = {
    'face_occluder': {
        'url': 'https://socaityfiles.blob.core.windows.net/model-zoo/face2face/face_occluder.onnx',
        'path': f'{MODELS_DIR}/face_occluder.onnx'
    }
    #'face_parser': {
    #    'url': 'https://github.com/facefusion/facefusion-assets/releases/download/models/face_parser.onnx',
    #    'path': f'{MODELS_DIR}/face_parser.onnx'
    #}
}



import os
import numpy as np
from typing import Union, Dict
from io import BytesIO

from face2face.core.compatibility.Face import Face
from face2face.core.modules.storage.file_writable_face import FileWriteableFace
from face2face.core.modules.utils import get_files_in_dir


def load_reference_face_from_file(face_embedding_file_path: Union[str, BytesIO]) -> Face:
    """
    Load a reference face from a file or BytesIO buffer.
    
    Args:
        face_embedding_file_path: Source of the face embedding. Can be:
            - str: Path to a .npy file containing the face embedding
            - BytesIO: Buffer containing the face embedding data
    
    Returns:
        Face: The loaded face object with its embedding and attributes
    
    Raises:
        ValueError: If the file is not found (for string paths) or cannot be loaded
        Exception: If the face data is invalid or cannot be converted to a Face object
    """
    if isinstance(face_embedding_file_path, str):
        if not os.path.exists(face_embedding_file_path):
            raise ValueError(f"Reference face {face_embedding_file_path} not found.")

    try:
        # note: potential security issue, if the file was not created with face2face
        embedding = np.load(face_embedding_file_path, allow_pickle=True)
        return FileWriteableFace.to_face(embedding)
    except Exception as e:
        raise f"Error loading reference face {face_embedding_file_path}: {e}"


def load_reference_faces_from_folder(folder_path: str) -> Dict[str, Face]:
    """
    Load all reference faces from a folder containing .npy files.
    
    Args:
        folder_path: Path to the folder containing face embedding files (.npy)
    
    Returns:
        Dict[str, Face]: Dictionary mapping face names (derived from filenames) to their Face objects
    
    Note:
        Only loads .npy files from the specified folder. Skips any files that cannot be loaded.
    """
    files = get_files_in_dir(folder_path, [".npy"])
    reference_faces = {}
    for file in files:
        face_name = os.path.basename(file)[:-4]
        embedding = load_reference_face_from_file(file)
        if embedding is None:
            continue
        reference_faces[face_name] = embedding

    return reference_faces




# avoid circular dependency but provide type hints
from __future__ import annotations
from typing import TYPE_CHECKING, Union, List, Dict

from media_toolkit.utils.generator_wrapper import SimpleGeneratorWrapper

if TYPE_CHECKING:
    from face2face.core.face2face import Face2Face

# other imports
from face2face.core.compatibility.Face import Face
from media_toolkit import VideoFile




class _Video_Swap:

    def swap_video(
            self: Face2Face,
            video: Union[str, VideoFile],
            faces: Union[str, dict, list, List[Face], Face],
            enhance_face_model: Union[str, None] = None,
            include_audio: bool = True
    ):
        """
        Swaps the faces in the video.
        :param video: the video to swap the faces in
        :param faces: the faces to swap in the video
        :param enhance_face_model: the face enhancement model to use. Use None for no enhancement
        :param include_audio: if True, the audio will be included in the output video
        """
        video = VideoFile().from_any(video)
        if isinstance(faces, dict):
            return self.swap_pairs_in_video(
                swap_pairs=faces, video=video, enhance_face_model=enhance_face_model, include_audio=include_audio
            )
        elif type(faces) in [list, str, Face]:
            return self.swap_to_face_in_video(
                faces=faces, video=video, enhance_face_model=enhance_face_model, include_audio=include_audio
            )

        raise NotImplementedError

    def swap_to_face_in_video(
            self: Face2Face,
            faces: str,
            video: Union[str, VideoFile],
            include_audio: bool = True,
            enhance_face_model: Union[str, None] = None
        ):
        """
        Swaps the face of the target video to the face of the reference image.
        :param face_name: the name of the reference face embedding
        :param video: the target video. Path to the file or VideoFile object
        :param include_audio: if True, the audio will be included in the output video
        :param enhance_face_model: the face enhancement model to use. Use None for no enhancement
        """
        video = VideoFile().from_any(video)
        if not isinstance(video, VideoFile):
            raise ValueError("Video must be a path or a VideoFile object")

        gen = video.to_video_stream(include_audio=include_audio)
        swap_gen = self.swap_to_face_generator(faces=faces, image_generator=gen, enhance_face_model=enhance_face_model)

        # allow tqdm to show better progress bar
        streamer = SimpleGeneratorWrapper(swap_gen, length=video.frame_count)

        new_video = VideoFile().from_video_stream(
            video_audio_stream=streamer,
            frame_rate=video.frame_rate,
            audio_sample_rate=video.audio_sample_rate
        )
        return new_video

    def swap_pairs_in_video(
            self: Face2Face,
            swap_pairs: dict,
            video: Union[str, VideoFile],
            include_audio: bool = True,
            enhance_face_model: Union[str, None] = None,
            recognition_threshold: float = 0.5
    ):
        """
        Swaps the reference faces in the target video.
        :param swap_pairs: a dict with the structure {source_face_name: target_face_name}
        :param video: the target video. Path to the file or VideoFile object
        :param enhance_face_model: the face enhancement model to use. Use None for no enhancement
        :param recognition_threshold: the threshold for face-recognition. Lower value -> more false positives
        :param include_audio: if True, the audio will be included in the output video
        """
        video = VideoFile().from_any(video)
        if not isinstance(video, VideoFile):
            raise ValueError("Video must be a path or a VideoFile object")

        gen = video.to_video_stream(include_audio=include_audio)

        swapper_gen = self.swap_pairs_generator(
                swap_pairs=swap_pairs,
                image_generator=gen,
                enhance_face_model=enhance_face_model,
                recognition_threshold=recognition_threshold
        )

        # allow tqdm to show better progress bar
        streamer = SimpleGeneratorWrapper(swapper_gen, length=video.frame_count)

        new_video = VideoFile().from_video_stream(
            video_audio_stream=streamer,
            frame_rate=video.frame_rate,
            audio_sample_rate=video.audio_sample_rate
        )
        return new_video



# -*- coding: utf-8 -*-
# @Organization  : insightface.ai
# @Author        : Jia Guo
# @Time          : 2021-05-04
# @Function      :


from __future__ import division

import glob
import os.path as osp

import numpy as np
import onnxruntime

__all__ = ['FaceAnalysis']

from face2face.core.compatibility.ArcFaceONNX import ArcFaceONNX
from face2face.core.compatibility.Attribute import Attribute
from face2face.core.compatibility.Face import Face
from face2face.core.compatibility.Landmark import Landmark
from face2face.core.compatibility.retinaface import RetinaFace



class FaceAnalysis:
    def __init__(self, model_dir: str, providers=None):

        onnxruntime.set_default_logger_severity(3)

        # loading all the insightface buffalo models and storing them in a dictionary
        self.models = {}
        onnx_files = glob.glob(osp.join(model_dir, '*.onnx'))
        onnx_files = sorted(onnx_files)
        for onnx_file in onnx_files:
            model = self._load_model(onnx_file, providers)
            self.models[model.taskname] = model

        self.det_model = self.models['detection']

    def _load_model(self, onnx_file: str, providers):
        if "1k3d68" in onnx_file or "2d106det" in onnx_file:
            return Landmark(model_file=onnx_file, providers=providers)
        if "det_10g" in onnx_file:
            return RetinaFace(model_file=onnx_file, providers=providers)
        if "genderage" in onnx_file:
            return Attribute(model_file=onnx_file, providers=providers)
        if "w600k_r50" in onnx_file:
            return ArcFaceONNX(model_file=onnx_file, providers=providers)

    def prepare(self, ctx_id, det_thresh=0.5, det_size=(640, 640)):
        self.det_thresh = det_thresh
        assert det_size is not None
        print('set det-size:', det_size)
        self.det_size = det_size
        for taskname, model in self.models.items():
            if taskname=='detection':
                model.prepare(ctx_id, input_size=det_size, det_thresh=det_thresh)
            else:
                model.prepare(ctx_id)

    def get(self, img, max_num=0):
        bboxes, kpss = self.det_model.detect(img,
                                             max_num=max_num,
                                             metric='default')
        if bboxes.shape[0] == 0:
            return []
        ret = []
        for i in range(bboxes.shape[0]):
            bbox = bboxes[i, 0:4]
            det_score = bboxes[i, 4]
            kps = None
            if kpss is not None:
                kps = kpss[i]
            face = Face(bbox=bbox, kps=kps, det_score=det_score)
            for taskname, model in self.models.items():
                if taskname=='detection':
                    continue
                model.get(img, face)
            ret.append(face)
        return ret

    def draw_on(self, img, faces):
        import cv2
        dimg = img.copy()
        for i in range(len(faces)):
            face = faces[i]
            box = face.bbox.astype(np.int)
            color = (0, 0, 255)
            cv2.rectangle(dimg, (box[0], box[1]), (box[2], box[3]), color, 2)
            if face.kps is not None:
                kps = face.kps.astype(np.int)
                #print(landmark.shape)
                for l in range(kps.shape[0]):
                    color = (0, 0, 255)
                    if l == 0 or l == 3:
                        color = (0, 255, 0)
                    cv2.circle(dimg, (kps[l][0], kps[l][1]), 1, color,
                               2)
            if face.gender is not None and face.age is not None:
                cv2.putText(dimg,'%s,%d'%(face.sex,face.age), (box[0]-1, box[1]-4),cv2.FONT_HERSHEY_COMPLEX,0.7,(0,255,0),1)

        return dimg




import os
from typing import Literal, Union, List

from fast_task_api import FastTaskAPI, ImageFile, JobProgress, MediaFile, VideoFile, MediaList, MediaDict

import numpy as np


from face2face.core.face2face import Face2Face
from media_toolkit.utils.generator_wrapper import SimpleGeneratorWrapper

from face2face.settings import ALLOW_EMBEDDING_SAVE_ON_SERVER


f2f = Face2Face()
app = FastTaskAPI(
    title="Face2Face",
    summary="Swap faces from images and videos. Create face embeddings.",
    version="0.0.8",
    contact={
        "name": "SocAIty",
        "url": "https://github.com/SocAIty/face2face",
    }
)

FACE_ENHANCE_MODELS = Literal['', 'gpen_bfr_512', 'gpen_bfr_1024', 'gpen_bfr_2048']


@app.task_endpoint("/swap_img_to_img", queue_size=500)
def swap_img_to_img(source_img: ImageFile, target_img: ImageFile, enhance_face_model: FACE_ENHANCE_MODELS = 'gpen_bfr_512'):
    """
    Swap faces between two images.

    Args:
        source_img: Source image containing the face(s) to swap from
        target_img: Target image containing the face(s) to swap to
        enhance_face_model: Face enhancement model to use. Defaults to 'gpen_bfr_512'

    Returns:
        ImageFile: The resulting image with swapped faces
    """
    swapped_img = f2f.swap_img_to_img(np.array(source_img), np.array(target_img), enhance_face_model=enhance_face_model)
    return ImageFile(file_name="swapped_img.png").from_np_array(swapped_img)


@app.task_endpoint("/add_face", queue_size=500)
def add_face(
    face_name: Union[str, List[str]],
    image: ImageFile,
    save: bool = ALLOW_EMBEDDING_SAVE_ON_SERVER
):
    """
    Add one or multiple reference face(s) to the face swapper.
    
    Args:
        face_name: Name(s) for the reference face(s).
            - If a single string, creates one face embedding
            - If a list of strings, creates embeddings for each face from left to right in the image
        image: The image from which to extract the face(s).
            - ImageFile: Standard image file
        save: Whether to save the face embeddings to disk.
            Note: This is controlled by ALLOW_EMBEDDING_SAVE_ON_SERVER setting
    
    Returns:
        Union[MediaFile, MediaDict]:
            - For single face: MediaFile containing the face embedding
            - For multiple faces: MediaDict mapping face names to their embeddings
    
    Raises:
        ValueError: If no face name is provided or no faces are detected in the image
    """
    # don't save embeddings on the server if the setting is False
    # this is useful in "multi-user" scenarios or if the server is not supposed to store any data
    save = save and ALLOW_EMBEDDING_SAVE_ON_SERVER

    faces = f2f.add_face(face_name=face_name, image=image, save=save)

    if isinstance(faces, dict):
        return MediaDict({
            face_name: MediaFile(file_name=f"{face_name}.npy").from_bytesio(face.to_bytes_io())
            for face_name, face in faces.items()
        })

    if isinstance(faces, tuple):
        return MediaFile(file_name=f"{face_name}.npy").from_bytesio(faces[1].to_bytes_io())


@app.task_endpoint("/swap", queue_size=500)
def swap(
    job_progress: JobProgress,
    faces: Union[List[str], dict, MediaFile, MediaList],
    media: Union[ImageFile, VideoFile],
    enhance_face_model: FACE_ENHANCE_MODELS = 'gpen_bfr_512'
):
    """
    Swap faces in an image or video.

    Args:
        faces: The face(s) to swap to. Can be:
            - str: Name of a reference face
            - dict: Swap pairs with structure {source_face_name: target_face_name}
            - list: List of face names or Face embeddings
            - MediaFile: Single face embedding file
            - MediaList: Multiple face embedding files
        media: The image or video to swap faces in
        enhance_face_model: Face enhancement model to use. Defaults to 'gpen_bfr_512'

    Returns:
        Union[ImageFile, VideoFile]: The resulting media with swapped faces

    Raises:
        ValueError: If no faces are provided or media type is unsupported
    """
    if isinstance(media, VideoFile):
        return swap_video(job_progress=job_progress, faces=faces, target_video=media, enhance_face_model=enhance_face_model, include_audio=True)

    job_progress.set_status(message="Started swapping faces")
    swapped_media = f2f.swap(faces=faces, media=media, enhance_face_model=enhance_face_model)

    if isinstance(swapped_media, np.ndarray):
        return ImageFile(file_name="swapped.png").from_np_array(swapped_media)

    return swapped_media


@app.task_endpoint("/swap_video", queue_size=10)
def swap_video(
        job_progress: JobProgress,
        faces: Union[List[str], dict, MediaFile, MediaList],
        target_video: VideoFile,
        include_audio: bool = True,
        enhance_face_model: FACE_ENHANCE_MODELS = 'gpen_bfr_512'
):
    """
    Swap faces in a video file.
    
    Args:
        face_name: The face(s) to swap to. Can be:
            - str: Name of a reference face
            - list: List of face names or Face objects
            - MediaFile: Single face embedding file
            - MediaList: Multiple face embedding files
        target_video: The video to swap faces in
        include_audio: Whether to include audio in the output video
        enhance_face_model: Face enhancement model to use. Defaults to 'gpen_bfr_512'

    Returns:
        VideoFile: The resulting video with swapped faces

    Raises:
        ValueError: If no faces are provided or video cannot be processed
    """
    def video_stream_gen():
        # generator reads the video stream and swaps the faces frame by frame
        gen = target_video.to_video_stream(include_audio=include_audio)
        swap_gen = f2f.swap_to_face_generator(faces, gen, enhance_face_model=enhance_face_model)
        # Swap the images one by one
        for i, swapped_audio_tuple in enumerate(swap_gen):
            audio = None
            if include_audio and len(swapped_audio_tuple) == 2:
                swapped_img, audio = swapped_audio_tuple
            else:
                swapped_img = swapped_audio_tuple

            # update progress. Swapping is 90% of the work
            if target_video.frame_count:
                percent_converted = float(i / target_video.frame_count)
                percent_total = round(percent_converted * 0.9, 2)
                if percent_total > 0.9:  # this is case if the estimation of cv2 is smaller than the actual video size
                    percent_total = 0.9

            else:
                percent_total = 0.5  # arbitrary number
            job_progress.set_status(message=f"swapping frame {i}", progress=percent_total)

            # Yield the image
            if include_audio and audio is not None:
                yield swapped_img, audio
            else:
                yield swapped_img

    # allow tqdm to show better progress bar
    streamer = SimpleGeneratorWrapper(video_stream_gen(), length=target_video.frame_count)

    # Create video
    output_video = VideoFile().from_video_stream(
        video_audio_stream=streamer,
        frame_rate=target_video.frame_rate,
        audio_sample_rate=target_video.audio_sample_rate
    )
    return output_video


# start the server on provided port
if __name__ == "__main__":
    # setting port to 8020 (referenced as so in socaity sdk for face2face) if not specified differently
    port = int(os.environ.get("FTAPI_PORT", 8020))
    app.start(port=port)



# avoid circular dependency but provide type hints
from __future__ import annotations
from typing import TYPE_CHECKING, Union, List, Dict

if TYPE_CHECKING:
    from face2face.core.face2face import Face2Face

# normal imports
from face2face.core.modules.utils.utils import load_image
from collections import OrderedDict
from face2face.core.compatibility.Face import Face
import numpy as np


class _FaceRecognition:
    """
    Face recognition - Mixin. Provides functions for identification, and swapping of known reference faces.
    """
    def face_recognition(
            self: Face2Face,
            image: np.array,
            face_names: Union[str, List[str], None] = None,
            threshold: float = 0.5
    ) -> list:
        """
        Given an image the method recognizes the faces in the image.
        :param image: the image to recognize the faces in
        :param face_names: the reference faces to compare the detected faces to.
            If None all stored face_embeddings are used.
        :param threshold: the threshold distance to recognize a face
        :return: A list with tuples in form [(recognized_face_name, dist, detected_face_in_image), ... ]
        """
        image = load_image(image)
        detected_faces = self.detect_faces(image)

        # Load reference faces
        ref_faces = self.load_faces(face_names)
        ref_faces = self._to_single_face_embeddings(ref_faces)

        # Detect faces and calculate face distances to source reference faces
        face_distances = self.calculate_face_distances(detected_faces, ref_faces)

        # prepare target_face vector based on the most similar reference face for each detected face
        # The source faces vector will have same length as the detected faces
        recognized_faces = []
        for i, face in enumerate(detected_faces):
            # check if the face is recognized
            closest_face, dist = next(iter(face_distances[i].items()), (None, 1))
            if closest_face is not None and dist < threshold:
                recognized_faces.append((closest_face, dist, face))
        return recognized_faces

    def swap_pairs(
            self: Face2Face,
            image: Union[np.array, str],
            swap_pairs: dict,
            enhance_face_model: str = 'gpen_bfr_512',
            threshold: float = 0.5
    ):
        """
        Based on the swap_pairs, swap the source faces to the target faces if they are recognized.
        1. Identify all persons in an image, by calculating the cosine distance between the embeddings >= threshold.
        2. Call the swap function with the target faces.

        """
        image = load_image(image)

        # recognize the faces of first swap partners in the image
        recognized_partner_faces = self.face_recognition(
            image,
            face_names=list(swap_pairs.keys()),
            threshold=threshold
        )

        # load the faces of the second swap partners
        swap_partner_embeddings = self.load_faces(face_names=list(swap_pairs.values()))
        swap_partner_embeddings = self._to_single_face_embeddings(swap_partner_embeddings)

        # swap faces of recognized partners to the second swap partners
        swap_embeddings = []
        for rec in recognized_partner_faces:
            partner = swap_pairs[rec[0]]
            swap_embeddings.append(swap_partner_embeddings[partner])

        return self._swap_faces(
            source_faces=swap_embeddings,
            target_faces=[rec[2] for rec in recognized_partner_faces],
            image=image,
            enhance_face_model=enhance_face_model
        )

    def swap_pairs_generator(
            self: Face2Face,
            swap_pairs: dict,
            image_generator,
            enhance_face_model: Union[str, None] = 'gpen_bfr_2048',
            recognition_threshold: float = 0.5
    ):
        """
        Swaps the reference faces in the target image.
        :param swap_pairs: a dict with the structure {source_face_name: target_face_name}
        :param image_generator: a generator that yields images in BGR format (read with cv2).
            Or a video_stream that yields (image, audio) like in media_toolkit.
        :return: a generator that yields the swapped images or tuples (image, audio)
        """
        if not isinstance(swap_pairs, dict):
            raise ValueError("Please provide a dict with the structure {source_face_name: target_face_name}")

        for i, target_image in enumerate(image_generator):
            # check if generator yields tuples (video, audio) or only images
            audio = None
            if isinstance(target_image, tuple) and len(target_image) == 2:
                target_image, audio = target_image

            try:
                swapped = self.swap_pairs(
                    swap_pairs=swap_pairs,
                    image=target_image,
                    enhance_face_model=enhance_face_model,
                    threshold=recognition_threshold
                )

                if audio is not None:
                    yield swapped, audio
                    continue

                yield swapped
                continue
            except Exception as e:
                print(f"Error in swapping frame {i}: {e}. Skipping image")
                if audio is not None:
                    yield target_image, audio
                    continue

                yield np.array(target_image)

    def calculate_face_distances(
            self: Face2Face,
            face_list_1: List[Face],
            face_list_2: Union[List[Face], Dict[str, Face]]
    ) -> list:
        """
        Calculate the face distances between the detected faces and the reference faces.
        :param face_list_1: the detected faces in the image
        :param face_list_2:
            the reference faces to calculate the distance to. If list, the faces are enumerated to dict with index as key.
            If in a loaded face embedding there are multiple faces; only the first face is used.
        :return: the face distances as a list of ordered and distance sorted dictionaries in form of
            [{face_1_0_to_face_2_0: distance, face_1_0_to_face_2_1: distance}, {face_1_1_to_face_2_0: distance, ...}  ...]
        """
        if len(face_list_1) == 0 or len(face_list_2) == 0:
            return {}

        if isinstance(face_list_2, list):
            face_list_2 = dict(enumerate(face_list_2))

        # flatten embeddings with multiple faces
        face_list_2_flat = self._to_single_face_embeddings(face_list_2)

        # Calculate distances
        face_distances = []  # [{face1_0: distance, face2_0: distance ..}, ...]
        for i, face in enumerate(face_list_1):
            face_dists = {}
            for reference_face_name, reference_face in face_list_2_flat.items():
                dist = self.calc_face_distance(face, reference_face)
                face_dists[reference_face_name] = dist
            face_dists = OrderedDict(sorted(face_dists.items(), key=lambda x: x[1], reverse=False))
            face_distances.append(face_dists)
        return face_distances

    @staticmethod
    def _to_single_face_embeddings(embeddings: dict) -> dict:
        """
        An embedding can store multiple faces. This function loads only the first face.
        :param embeddings: the embeddings
        :return: the { faces: first_face_embedding }
        """
        return {
            face_name: face[0] if isinstance(face, list) else face
            for face_name, face in embeddings.items()
        }

    @staticmethod
    def calc_face_distance(face: Face, reference_face: Face) -> float:
        if hasattr(face, 'normed_embedding') and hasattr(reference_face, 'normed_embedding'):
            return 1 - np.dot(face.normed_embedding, reference_face.normed_embedding)
        return 1


    # @staticmethod
    # def cosine_distance(embedding1, embedding2):
    #    """
    #    Calculate the cosine distance between two embeddings.
    #    :param embedding1: the first embedding
    #    :param embedding2: the second embedding
    #    :return: the cosine distance
    #    """
    #    return np.dot(embedding1, embedding2) / (np.linalg.norm(embedding1) * np.linalg.norm(embedding2))



import threading
from typing import Tuple

import cv2
import onnxruntime
from cv2.typing import Size
import numpy as np

from face2face.core.modules.face_enhance.face_enhance_models import get_model_config, WARP_TEMPLATES
from face2face.settings import EXECUTION_PROVIDER, DEVICE_ID
from face2face.core.compatibility.Face import Face

# Thread Lock required for onnx engine
THREAD_LOCK = threading.Lock()


def create_static_box_mask(
        crop_size: Size,
        face_mask_blur: float,
        face_mask_padding: Tuple[int, int, int, int]
) -> np.array:
    blur_amount = int(crop_size[0] * 0.5 * face_mask_blur)
    blur_area = max(blur_amount // 2, 1)
    box_mask = np.ones(crop_size, np.float32)
    box_mask[:max(blur_area, int(crop_size[1] * face_mask_padding[0] / 100)), :] = 0
    box_mask[-max(blur_area, int(crop_size[1] * face_mask_padding[2] / 100)):, :] = 0
    box_mask[:, :max(blur_area, int(crop_size[0] * face_mask_padding[3] / 100))] = 0
    box_mask[:, -max(blur_area, int(crop_size[0] * face_mask_padding[1] / 100)):] = 0
    if blur_amount > 0:
        box_mask = cv2.GaussianBlur(box_mask, (0, 0), blur_amount * 0.25)
    return box_mask


def prepare_crop_frame(crop_vision_frame: np.array) -> np.array:
    crop_vision_frame = crop_vision_frame[:, :, ::-1] / 255.0
    crop_vision_frame = (crop_vision_frame - 0.5) / 0.5
    crop_vision_frame = np.expand_dims(crop_vision_frame.transpose(2, 0, 1), axis=0).astype(np.float32)
    return crop_vision_frame


def estimate_matrix_by_face_landmark_5(
        face_landmark_5: np.array,
        warp_template: str,
        crop_size: Size
) -> np.array:
    normed_warp_template = WARP_TEMPLATES.get(warp_template) * crop_size
    affine_matrix = cv2.estimateAffinePartial2D(
            face_landmark_5, normed_warp_template, method=cv2.RANSAC, ransacReprojThreshold=100
    )[0]
    return affine_matrix


def warp_face_by_face_landmark_5(
        temp_vision_frame: np.array,
        face_landmark_5: np.array,
        warp_template: str,  # the ones defined in model templates
        crop_size: Size
) -> Tuple[np.array, np.array]:
    affine_matrix = estimate_matrix_by_face_landmark_5(face_landmark_5, warp_template, crop_size)
    crop_vision_frame = cv2.warpAffine(temp_vision_frame, affine_matrix, crop_size, borderMode=cv2.BORDER_REPLICATE,
                                       flags=cv2.INTER_AREA)
    return crop_vision_frame, affine_matrix


def normalize_crop_frame(crop_vision_frame: np.array) -> np.array:
    crop_vision_frame = np.clip(crop_vision_frame, -1, 1)
    crop_vision_frame = (crop_vision_frame + 1) / 2
    crop_vision_frame = crop_vision_frame.transpose(1, 2, 0)
    crop_vision_frame = (crop_vision_frame * 255.0).round()
    crop_vision_frame = crop_vision_frame.astype(np.uint8)[:, :, ::-1]
    return crop_vision_frame


def get_onnx_inference_session(
        model_path: str,
        execution_provider: tuple = None  # tuple in form ov (execution_provider_name, settings)
):
    if execution_provider is None:
        providers = [(EXECUTION_PROVIDER, {"device_id": DEVICE_ID})]

    with THREAD_LOCK:
        session = onnxruntime.InferenceSession(model_path, providers=providers)
    return session


def apply_enhance(crop_vision_frame: np.array, model_path: str) -> np.array:
    inference_session = get_onnx_inference_session(model_path=model_path)
    frame_processor_inputs = {}
    for frame_processor_input in inference_session.get_inputs():
        if frame_processor_input.name == 'input':
            frame_processor_inputs[frame_processor_input.name] = crop_vision_frame
        if frame_processor_input.name == 'weight':
            weight = np.array([1]).astype(np.double)
            frame_processor_inputs[frame_processor_input.name] = weight

    with THREAD_LOCK:
        crop_vision_frame = inference_session.run(None, frame_processor_inputs)[0][0]

    return crop_vision_frame


def paste_back(
        temp_vision_frame: np.array,
        crop_vision_frame: np.array,
        crop_mask: np.array,
        affine_matrix: np.array
) -> np.array:
    inverse_matrix = cv2.invertAffineTransform(affine_matrix)
    temp_size = temp_vision_frame.shape[:2][::-1]
    inverse_mask = cv2.warpAffine(crop_mask, inverse_matrix, temp_size).clip(0, 1)
    inverse_vision_frame = cv2.warpAffine(crop_vision_frame, inverse_matrix, temp_size, borderMode=cv2.BORDER_REPLICATE)
    paste_vision_frame = temp_vision_frame.copy()
    paste_vision_frame[:, :, 0] = inverse_mask * inverse_vision_frame[:, :, 0] + (1 - inverse_mask) * temp_vision_frame[
                                                                                                      :, :, 0]
    paste_vision_frame[:, :, 1] = inverse_mask * inverse_vision_frame[:, :, 1] + (1 - inverse_mask) * temp_vision_frame[
                                                                                                      :, :, 1]
    paste_vision_frame[:, :, 2] = inverse_mask * inverse_vision_frame[:, :, 2] + (1 - inverse_mask) * temp_vision_frame[
                                                                                                      :, :, 2]
    return paste_vision_frame


def blend_frame(
        temp_vision_frame: np.array,
        paste_vision_frame: np.array,
        blend_strength: float = 0.5  # value between 0 and 1
) -> np.array:
    face_enhancer_blend = 1 - blend_strength
    temp_vision_frame = cv2.addWeighted(
        temp_vision_frame, face_enhancer_blend, paste_vision_frame, 1 - face_enhancer_blend, 0
    )
    return temp_vision_frame


def enhance_face(
        target_face: Face,
        temp_vision_frame: np.array,
        model='gfpgan_1.4'
) -> np.array:
    config = get_model_config(model)
    model_template = config.get('template')
    model_size = config.get('size')
    model_path = config.get('path')

    landmark = target_face.get("kps")  # get('landmark_2d_106')

    crop_vision_frame, affine_matrix = warp_face_by_face_landmark_5(
        temp_vision_frame,
        landmark,
        model_template,
        model_size
    )
    box_mask = create_static_box_mask(
        crop_vision_frame.shape[:2][::-1],
        face_mask_blur=0.0,  # (0, 0, 0, 0),  # facefusion.globals.face_mask_blur,
        face_mask_padding=(0, 0, 0, 0)
    )
    crop_mask_list = [box_mask]

    # if 'occlusion' in facefusion.globals.face_mask_types:
    #    occlusion_mask = create_occlusion_mask(crop_vision_frame)
    #    crop_mask_list.append(occlusion_mask)

    crop_vision_frame = prepare_crop_frame(crop_vision_frame)
    crop_vision_frame = apply_enhance(crop_vision_frame=crop_vision_frame, model_path=model_path)
    crop_vision_frame = normalize_crop_frame(crop_vision_frame)
    crop_mask = np.minimum.reduce(crop_mask_list).clip(0, 1)
    paste_vision_frame = paste_back(temp_vision_frame, crop_vision_frame, crop_mask, affine_matrix)
    temp_vision_frame = blend_frame(temp_vision_frame, paste_vision_frame)
    return temp_vision_frame
